@@ -25,6 +25,16 @@
 
 G_BEGIN_DECLS
 
+GString* udisks_string_concat (GString*, GString*);
+void udisks_string_wipe_and_free (GString*);
+
+gboolean udisks_variant_lookup_binary (GVariant     *dict,
+                                       const gchar  *name,
+                                       GString     **contents);
+
+gboolean udisks_variant_get_binary (GVariant  *variant,
+                                    GString  **contents);
+
 gchar *udisks_decode_udev_string (const gchar *str);
 
 void udisks_safe_append_to_object_path (GString      *str,
@@ -55,6 +65,14 @@ gboolean udisks_daemon_util_check_authorization_sync (UDisksDaemon          *dae
                                                       const gchar           *message,
                                                       GDBusMethodInvocation *invocation);
 
+gboolean udisks_daemon_util_check_authorization_sync_with_error (UDisksDaemon           *daemon,
+                                                                 UDisksObject           *object,
+                                                                 const gchar            *action_id,
+                                                                 GVariant               *options,
+                                                                 const gchar            *message,
+                                                                 GDBusMethodInvocation  *invocation,
+                                                                 GError                **error);
+
 gboolean udisks_daemon_util_get_caller_uid_sync (UDisksDaemon            *daemon,
                                                  GDBusMethodInvocation   *invocation,
                                                  GCancellable            *cancellable,
@@ -84,13 +102,30 @@ gboolean udisks_daemon_util_file_set_contents (const gchar  *filename,
                                                gint          mode_for_new_file,
                                                GError      **error);
 
-UDisksInhibitCookie *udisks_daemon_util_inhibit_system_sync   (const gchar          *reason);
+UDisksInhibitCookie *udisks_daemon_util_inhibit_system_sync   (const gchar            *reason);
 void                 udisks_daemon_util_uninhibit_system_sync (UDisksInhibitCookie  *cookie);
 
 gchar *udisks_daemon_util_get_free_mdraid_device (void);
 
 guint16 udisks_ata_identify_get_word (const guchar *identify_data, guint word_number);
 
+/* Utility macro for policy verification. */
+#define UDISKS_DAEMON_CHECK_AUTHORIZATION(daemon,                   \
+                                          object,                   \
+                                          action_id,                \
+                                          options,                  \
+                                          message,                  \
+                                          invocation)               \
+  if (! udisks_daemon_util_check_authorization_sync ((daemon),      \
+                                                     (object),      \
+                                                     (action_id),   \
+                                                     (options),     \
+                                                     (message),     \
+                                                     (invocation))) \
+    { \
+      goto out; \
+    }
+
 G_END_DECLS
 
 #endif /* __UDISKS_DAEMON_UTIL_H__ */
