@@ -29,7 +29,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <mntent.h>
-#include <sys/types.h>
+#include <sys/sysmacros.h>
 #ifdef HAVE_ACL
 #include <sys/acl.h>
 #endif
@@ -216,7 +216,7 @@ is_in_filesystem_file (const gchar *filesystems_file,
                       error->message,
                       g_quark_to_string (error->domain),
                       error->code);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -293,7 +293,7 @@ static const gchar *vfat_allow_gid_self[] = { "gid=", NULL };
 /* this is assuming that ntfs-3g is used */
 
 static const gchar *ntfs_defaults[] = { "uid=", "gid=", NULL };
-static const gchar *ntfs_allow[] = { "umask=", "dmask=", "fmask=", "locale=", "norecover", "ignore_case", "windows_names", "compression", "nocompression", NULL };
+static const gchar *ntfs_allow[] = { "umask=", "dmask=", "fmask=", "locale=", "norecover", "ignore_case", "windows_names", "compression", "nocompression", "big_writes", NULL };
 static const gchar *ntfs_allow_uid_self[] = { "uid=", NULL };
 static const gchar *ntfs_allow_gid_self[] = { "gid=", NULL };
 
@@ -313,11 +313,18 @@ static const gchar *udf_allow_gid_self[] = { "gid=", NULL };
 
 /* ---------------------- exfat -------------------- */
 
-static const gchar *exfat_defaults[] = { "uid=", "gid=", "iocharset=utf8", "namecase=0", "errors=remount-ro", "umask=0077", NULL };
+static const gchar *exfat_defaults[] = { "uid=", "gid=", "iocharset=utf8", "namecase=0", "errors=remount-ro", NULL };
 static const gchar *exfat_allow[] = { "dmask=", "errors=", "fmask=", "iocharset=", "namecase=", "umask=", NULL };
 static const gchar *exfat_allow_uid_self[] = { "uid=", NULL };
 static const gchar *exfat_allow_gid_self[] = { "gid=", NULL };
 
+/* ---------------------- hfs+ -------------------- */
+
+static const gchar *hfsplus_defaults[] = { "uid=", "gid=", "nls=utf8", NULL };
+static const gchar *hfsplus_allow[] = { "creator=", "type=", "umask=", "session=", "part=", "decompose", "nodecompose", "force", "nls=", NULL };
+static const gchar *hfsplus_allow_uid_self[] = { "uid=", NULL };
+static const gchar *hfsplus_allow_gid_self[] = { "gid=", NULL };
+
 /* ------------------------------------------------ */
 /* TODO: support context= */
 
@@ -330,6 +337,7 @@ static const FSMountOptions fs_mount_options[] =
     { "iso9660", iso9660_defaults, iso9660_allow, iso9660_allow_uid_self, iso9660_allow_gid_self },
     { "udf", udf_defaults, udf_allow, udf_allow_uid_self, udf_allow_gid_self },
     { "exfat", exfat_defaults, exfat_allow, exfat_allow_uid_self, exfat_allow_gid_self },
+    { "hfsplus", hfsplus_defaults, hfsplus_allow, hfsplus_allow_uid_self, hfsplus_allow_gid_self },
   };
 
 /* ------------------------------------------------ */
@@ -630,9 +638,9 @@ subst_str_and_escape (const gchar *str,
  * Returns: A valid UTF-8 string with the filesystem type (may be "auto") or %NULL if @error is set. Free with g_free().
  */
 static gchar *
-calculate_fs_type (UDisksBlock               *block,
-                   GVariant                  *given_options,
-                   GError                   **error)
+calculate_fs_type (UDisksBlock  *block,
+                   GVariant     *given_options,
+                   GError      **error)
 {
   gchar *fs_type_to_use = NULL;
   const gchar *probed_fs_type = NULL;
@@ -711,12 +719,12 @@ calculate_fs_type (UDisksBlock               *block,
  * Returns: A string with mount options or %NULL if @error is set. Free with g_free().
  */
 static gchar *
-calculate_mount_options (UDisksDaemon              *daemon,
-                         UDisksBlock               *block,
-                         uid_t                      caller_uid,
-                         const gchar               *fs_type,
-                         GVariant                  *options,
-                         GError                   **error)
+calculate_mount_options (UDisksDaemon  *daemon,
+                         UDisksBlock   *block,
+                         uid_t          caller_uid,
+                         const gchar   *fs_type,
+                         GVariant      *options,
+                         GError       **error)
 {
   const FSMountOptions *fsmo;
   gchar **options_to_use;
@@ -855,13 +863,13 @@ add_acl (const gchar  *path,
  * Returns: A UTF-8 string with the mount point to use or %NULL if @error is set. Free with g_free().
  */
 static gchar *
-calculate_mount_point (UDisksDaemon              *daemon,
-                       UDisksBlock               *block,
-                       uid_t                      uid,
-                       gid_t                      gid,
-                       const gchar               *user_name,
-                       const gchar               *fs_type,
-                       GError                   **error)
+calculate_mount_point (UDisksDaemon  *daemon,
+                       UDisksBlock   *block,
+                       uid_t          uid,
+                       gid_t          gid,
+                       const gchar   *user_name,
+                       const gchar   *fs_type,
+                       GError       **error)
 {
   UDisksLinuxBlockObject *object = NULL;
   gboolean fs_shared = FALSE;
@@ -1043,10 +1051,10 @@ has_option (const gchar *options,
 }
 
 static gboolean
-is_in_fstab (UDisksBlock        *block,
-             const gchar        *fstab_path,
-             gchar             **out_mount_point,
-             gchar             **out_mount_options)
+is_in_fstab (UDisksBlock  *block,
+             const gchar  *fstab_path,
+             gchar       **out_mount_point,
+             gchar       **out_mount_options)
 {
   gboolean ret;
   FILE *f;
@@ -1132,9 +1140,9 @@ is_in_fstab (UDisksBlock        *block,
  * TODO: check if systemd has a specific "unit" for the device
  */
 static gboolean
-is_system_managed (UDisksBlock        *block,
-                   gchar             **out_mount_point,
-                   gchar             **out_mount_options)
+is_system_managed (UDisksBlock  *block,
+                   gchar       **out_mount_point,
+                   gchar       **out_mount_options)
 {
   gboolean ret;
 
@@ -1154,9 +1162,9 @@ is_system_managed (UDisksBlock        *block,
 
 /* runs in thread dedicated to handling @invocation */
 static gboolean
-handle_mount (UDisksFilesystem       *filesystem,
-              GDBusMethodInvocation  *invocation,
-              GVariant               *options)
+handle_mount (UDisksFilesystem      *filesystem,
+              GDBusMethodInvocation *invocation,
+              GVariant              *options)
 {
   UDisksObject *object;
   UDisksBlock *block;
@@ -1247,7 +1255,7 @@ handle_mount (UDisksFilesystem       *filesystem,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1397,7 +1405,7 @@ handle_mount (UDisksFilesystem       *filesystem,
   if (fs_type_to_use == NULL)
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1412,7 +1420,7 @@ handle_mount (UDisksFilesystem       *filesystem,
   if (mount_options_to_use == NULL)
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1460,7 +1468,7 @@ handle_mount (UDisksFilesystem       *filesystem,
   if (mount_point_to_use == NULL)
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1557,9 +1565,9 @@ get_error_code_for_umount (gint         exit_status,
 
 /* runs in thread dedicated to handling @invocation */
 static gboolean
-handle_unmount (UDisksFilesystem       *filesystem,
-                GDBusMethodInvocation  *invocation,
-                GVariant               *options)
+handle_unmount (UDisksFilesystem      *filesystem,
+                GDBusMethodInvocation *invocation,
+                GVariant              *options)
 {
   UDisksObject *object;
   UDisksBlock *block;
@@ -1625,7 +1633,7 @@ handle_unmount (UDisksFilesystem       *filesystem,
   if (!udisks_daemon_util_get_caller_uid_sync (daemon, invocation, NULL, &caller_uid, NULL, NULL, &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1809,33 +1817,12 @@ handle_unmount (UDisksFilesystem       *filesystem,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
-static void
-on_set_label_job_completed (UDisksJob   *job,
-                            gboolean     success,
-                            const gchar *message,
-                            gpointer     user_data)
-{
-  GDBusMethodInvocation *invocation = G_DBUS_METHOD_INVOCATION (user_data);
-  UDisksFilesystem *filesystem;
-
-  filesystem = UDISKS_FILESYSTEM (g_dbus_method_invocation_get_user_data (invocation));
-
-  if (success)
-    udisks_filesystem_complete_set_label (filesystem, invocation);
-  else
-    g_dbus_method_invocation_return_error (invocation,
-                                           UDISKS_ERROR,
-                                           UDISKS_ERROR_FAILED,
-                                           "Error setting label: %s",
-                                           message);
-}
-
 /* runs in thread dedicated to handling method call */
 static gboolean
-handle_set_label (UDisksFilesystem       *filesystem,
-                  GDBusMethodInvocation  *invocation,
-                  const gchar            *label,
-                  GVariant               *options)
+handle_set_label (UDisksFilesystem      *filesystem,
+                  GDBusMethodInvocation *invocation,
+                  const gchar           *label,
+                  GVariant              *options)
 {
   UDisksBlock *block;
   UDisksObject *object;
@@ -1843,13 +1830,15 @@ handle_set_label (UDisksFilesystem       *filesystem,
   const gchar *probed_fs_usage;
   const gchar *probed_fs_type;
   const FSInfo *fs_info;
-  UDisksBaseJob *job;
   const gchar *action_id;
   const gchar *message;
   gchar *real_label = NULL;
   uid_t caller_uid;
   gid_t caller_gid;
   gchar *command;
+  gint status = 0;
+  gchar *out_message = NULL;
+  gboolean success = FALSE;
   gchar *tmp;
   GError *error;
 
@@ -1878,7 +1867,7 @@ handle_set_label (UDisksFilesystem       *filesystem,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1995,24 +1984,39 @@ handle_set_label (UDisksFilesystem       *filesystem,
       g_free (tmp);
     }
 
-  job = udisks_daemon_launch_spawned_job (daemon,
-                                          object,
-                                          "filesystem-modify", caller_uid,
-                                          NULL, /* cancellable */
-                                          0,    /* uid_t run_as_uid */
-                                          0,    /* uid_t run_as_euid */
-                                          NULL, /* input_string */
-                                          "%s", command);
-  g_signal_connect (job,
-                    "completed",
-                    G_CALLBACK (on_set_label_job_completed),
-                    invocation);
+  success = udisks_daemon_launch_spawned_job_sync (daemon,
+                                                   object,
+                                                   "filesystem-modify", caller_uid,
+                                                   NULL, /* cancellable */
+                                                   0,    /* uid_t run_as_uid */
+                                                   0,    /* uid_t run_as_euid */
+                                                   &status,
+                                                   &out_message,
+                                                   NULL, /* input_string */
+                                                   "%s", command);
+
+  /* XXX: label property is automatically updated after an udev change
+   * event for this device, but udev sometimes returns the old label
+   * so just trigger the uevent again now to be sure the property
+   * has been updated
+  */
+  udisks_linux_block_object_trigger_uevent (UDISKS_LINUX_BLOCK_OBJECT (object));
+
+  if (success)
+    udisks_filesystem_complete_set_label (filesystem, invocation);
+  else
+    g_dbus_method_invocation_return_error (invocation,
+                                           UDISKS_ERROR,
+                                           UDISKS_ERROR_FAILED,
+                                           "Error setting label: %s",
+                                           out_message);
 
  out:
   /* for some FSes we need to copy and modify label; free our copy */
   g_free (real_label);
   g_free (command);
   g_clear_object (&object);
+  g_free (out_message);
   return TRUE; /* returning TRUE means that we handled the method invocation */
 }
 
