@@ -40,6 +40,7 @@
 #include "udisksstate.h"
 #include "udisksdaemonutil.h"
 #include "udiskslinuxdevice.h"
+#include "udiskslinuxblock.h"
 
 /**
  * SECTION:udiskslinuxmdraid
@@ -255,6 +256,18 @@ member_cmpfunc (GVariant **a,
   return slot_a - slot_b;
 }
 
+static const char *
+sync_action_to_job_id (const char *sync_action)
+{
+  if (g_strcmp0 (sync_action, "check") == 0)
+    return "mdraid-check-job";
+  else if (g_strcmp0 (sync_action, "repair") == 0)
+    return "mdraid-repair-job";
+  else if (g_strcmp0 (sync_action, "recover") == 0)
+    return "mdraid-recover-job";
+  else
+    return "mdraid-sync-job";
+}
 
 /**
  * udisks_linux_mdraid_update:
@@ -291,6 +304,7 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
   UDisksDaemon *daemon = NULL;
   gboolean has_redundancy = FALSE;
   gboolean has_stripes = FALSE;
+  UDisksBaseJob *job = NULL;
 
   daemon = udisks_linux_mdraid_object_get_daemon (object);
 
@@ -312,17 +326,33 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
     {
       device = UDISKS_LINUX_DEVICE (member_devices->data);
       num_devices = g_udev_device_get_property_as_int (device->udev_device, "UDISKS_MD_MEMBER_DEVICES");
+      if (! num_devices)
+          num_devices = g_udev_device_get_property_as_int (device->udev_device, "STORAGED_MD_MEMBER_DEVICES");
       level = g_udev_device_get_property (device->udev_device, "UDISKS_MD_MEMBER_LEVEL");
+      if (! level)
+          level = g_udev_device_get_property (device->udev_device, "STORAGED_MD_MEMBER_LEVEL");
       uuid = g_udev_device_get_property (device->udev_device, "UDISKS_MD_MEMBER_UUID");
+      if (! uuid)
+          uuid = g_udev_device_get_property (device->udev_device, "STORAGED_MD_MEMBER_UUID");
       name = g_udev_device_get_property (device->udev_device, "UDISKS_MD_MEMBER_NAME");
+      if (! name)
+          name = g_udev_device_get_property (device->udev_device, "STORAGED_MD_MEMBER_NAME");
     }
   else
     {
       device = UDISKS_LINUX_DEVICE (raid_device);
       num_devices = g_udev_device_get_property_as_int (device->udev_device, "UDISKS_MD_DEVICES");
+      if (! num_devices)
+          num_devices = g_udev_device_get_property_as_int (device->udev_device, "STORAGED_MD_DEVICES");
       level = g_udev_device_get_property (device->udev_device, "UDISKS_MD_LEVEL");
+      if (! level)
+          level = g_udev_device_get_property (device->udev_device, "STORAGED_MD_LEVEL");
       uuid = g_udev_device_get_property (device->udev_device, "UDISKS_MD_UUID");
+      if (! uuid)
+          uuid = g_udev_device_get_property (device->udev_device, "STORAGED_MD_UUID");
       name = g_udev_device_get_property (device->udev_device, "UDISKS_MD_NAME");
+      if (! name)
+          name = g_udev_device_get_property (device->udev_device, "STORAGED_MD_NAME");
     }
 
   /* figure out size */
@@ -341,6 +371,8 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
   udisks_mdraid_set_num_devices (iface, num_devices);
   udisks_mdraid_set_size (iface, size);
 
+  udisks_mdraid_set_running (iface, raid_device != NULL);
+
   if (g_strcmp0 (level, "raid1") == 0 ||
       g_strcmp0 (level, "raid4") == 0 ||
       g_strcmp0 (level, "raid5") == 0 ||
@@ -401,6 +433,46 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
           sync_remaining_time = ((guint64) G_USEC_PER_SEC) * num_bytes_remaining / sync_rate;
         }
     }
+
+  if (sync_action)
+    {
+      if (g_strcmp0 (sync_action, "idle") != 0)
+        {
+          if (! udisks_linux_mdraid_object_has_sync_job (object))
+            {
+              /* Launch a job */
+              job = udisks_daemon_launch_simple_job (daemon,
+                                                     UDISKS_OBJECT (object),
+                                                     sync_action_to_job_id (sync_action),
+                                                     0,
+                                                     NULL /* cancellable */);
+
+              /* Mark the job as not cancellable. It simply has to finish... */
+              udisks_job_set_cancelable (UDISKS_JOB (job), FALSE);
+              udisks_linux_mdraid_object_set_sync_job (object, job);
+            }
+          else
+            job = udisks_linux_mdraid_object_get_sync_job (object);
+
+          /* Update the job's interface */
+          udisks_job_set_progress (UDISKS_JOB (job), sync_completed_val);
+          udisks_job_set_progress_valid (UDISKS_JOB (job), TRUE);
+          udisks_job_set_rate (UDISKS_JOB (job), sync_rate);
+
+          udisks_job_set_expected_end_time (UDISKS_JOB (job),
+                                            g_get_real_time () + sync_remaining_time);
+        }
+      else
+        {
+          if (udisks_linux_mdraid_object_has_sync_job (object))
+            {
+              /* Complete the job */
+              udisks_linux_mdraid_object_complete_sync_job (object,
+                                                            TRUE,
+                                                            "Finished");
+            }
+        }
+    }
   udisks_mdraid_set_sync_completed (iface, sync_completed_val);
   udisks_mdraid_set_sync_rate (iface, sync_rate);
   udisks_mdraid_set_sync_remaining_time (iface, sync_remaining_time);
@@ -527,6 +599,10 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
     }
   udisks_mdraid_set_active_devices (iface, g_variant_builder_end (&builder));
 
+  udisks_mdraid_set_child_configuration (iface,
+                                         udisks_linux_find_child_configuration (daemon,
+                                                                                uuid));
+
  out:
   g_free (sync_completed);
   g_free (sync_action);
@@ -618,7 +694,7 @@ handle_start (UDisksMDRaid           *_mdraid,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -745,9 +821,10 @@ handle_start (UDisksMDRaid           *_mdraid,
 /* ---------------------------------------------------------------------------------------------------- */
 
 static gboolean
-handle_stop (UDisksMDRaid           *_mdraid,
-             GDBusMethodInvocation  *invocation,
-             GVariant               *options)
+udisks_linux_mdraid_stop (UDisksMDRaid           *_mdraid,
+                          GDBusMethodInvocation  *invocation,
+                          GVariant               *options,
+                          GError                **error)
 {
   UDisksLinuxMDRaid *mdraid = UDISKS_LINUX_MDRAID (_mdraid);
   UDisksDaemon *daemon;
@@ -759,38 +836,39 @@ handle_stop (UDisksMDRaid           *_mdraid,
   UDisksLinuxDevice *raid_device = NULL;
   const gchar *device_file = NULL;
   gchar *escaped_device_file = NULL;
-  GError *error = NULL;
   gchar *error_message = NULL;
+  gboolean ret;
 
-  object = udisks_daemon_util_dup_object (mdraid, &error);
+  object = udisks_daemon_util_dup_object (mdraid, error);
   if (object == NULL)
     {
-      g_dbus_method_invocation_take_error (invocation, error);
+      ret = FALSE;
       goto out;
     }
 
   daemon = udisks_linux_mdraid_object_get_daemon (object);
   state = udisks_daemon_get_state (daemon);
 
-  error = NULL;
   if (!udisks_daemon_util_get_caller_uid_sync (daemon,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
                                                &caller_gid,
                                                NULL,
-                                               &error))
+                                               error))
     {
-      g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      ret = FALSE;
       goto out;
     }
 
   raid_device = udisks_linux_mdraid_object_get_device (object);
   if (raid_device == NULL)
     {
-      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
-                                             "RAID Array is not running");
+      g_set_error (error,
+                   UDISKS_ERROR,
+                   UDISKS_ERROR_FAILED,
+                   "RAID Array is not running");
+      ret = FALSE;
       goto out;
     }
 
@@ -813,13 +891,17 @@ handle_stop (UDisksMDRaid           *_mdraid,
       /* TODO: variables */
       message = N_("Authentication is required to stop a RAID array");
       action_id = "org.freedesktop.udisks2.manage-md-raid";
-      if (!udisks_daemon_util_check_authorization_sync (daemon,
-                                                        UDISKS_OBJECT (object),
-                                                        action_id,
-                                                        options,
-                                                        message,
-                                                        invocation))
-        goto out;
+      if (!udisks_daemon_util_check_authorization_sync_with_error (daemon,
+                                                                   UDISKS_OBJECT (object),
+                                                                   action_id,
+                                                                   options,
+                                                                   message,
+                                                                   invocation,
+                                                                   error))
+        {
+          ret = FALSE;
+          goto out;
+        }
     }
 
   device_file = g_udev_device_get_device_file (raid_device->udev_device);
@@ -837,23 +919,42 @@ handle_stop (UDisksMDRaid           *_mdraid,
                                               "mdadm --stop %s",
                                               escaped_device_file))
     {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Error stopping RAID array %s: %s",
-                                             device_file,
-                                             error_message);
+      g_set_error (error,
+                   UDISKS_ERROR,
+                   UDISKS_ERROR_FAILED,
+                   "Error stopping RAID array %s: %s",
+                   device_file,
+                   error_message);
+      ret = FALSE;
       goto out;
     }
 
-  udisks_mdraid_complete_stop (_mdraid, invocation);
+  ret = TRUE;
 
  out:
   g_free (error_message);
   g_free (escaped_device_file);
   g_clear_object (&raid_device);
   g_clear_object (&object);
-  return TRUE; /* returning TRUE means that we handled the method invocation */
+  return ret;
+}
+
+static gboolean
+handle_stop (UDisksMDRaid          *_mdraid,
+             GDBusMethodInvocation *invocation,
+             GVariant              *options)
+{
+  GError *error = NULL;
+
+  if (!udisks_linux_mdraid_stop (_mdraid,
+                                   invocation,
+                                   options,
+                                   &error))
+    g_dbus_method_invocation_take_error (invocation, error);
+  else
+    udisks_mdraid_complete_stop (_mdraid, invocation);
+
+  return TRUE;
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -966,7 +1067,7 @@ handle_remove_device (UDisksMDRaid           *_mdraid,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1169,7 +1270,7 @@ handle_add_device (UDisksMDRaid           *_mdraid,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1308,7 +1409,7 @@ handle_set_bitmap_location (UDisksMDRaid           *_mdraid,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1379,6 +1480,7 @@ handle_set_bitmap_location (UDisksMDRaid           *_mdraid,
     }
 
   udisks_mdraid_complete_add_device (_mdraid, invocation);
+  udisks_linux_mdraid_update (mdraid, object);
 
  out:
   g_free (error_message);
@@ -1430,7 +1532,7 @@ handle_request_sync_action (UDisksMDRaid           *_mdraid,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1514,6 +1616,190 @@ handle_request_sync_action (UDisksMDRaid           *_mdraid,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+static gboolean
+udisks_linux_mdraid_delete (UDisksMDRaid           *mdraid,
+                            GDBusMethodInvocation  *invocation,
+                            GVariant               *options,
+                            GError                **error)
+{
+  UDisksLinuxMDRaidObject *object;
+  UDisksDaemon *daemon;
+  uid_t caller_uid;
+  gid_t caller_gid;
+  const gchar *message;
+  const gchar *action_id;
+  gboolean teardown_flag = FALSE;
+  GList *member_devices = NULL;
+  GList *l;
+  UDisksLinuxDevice *raid_device = NULL;
+  gboolean ret;
+
+  g_variant_lookup (options, "tear-down", "b", &teardown_flag);
+
+  /* Delete is just stop followed by wiping of all members.
+   */
+
+  object = udisks_daemon_util_dup_object (mdraid, error);
+  if (object == NULL)
+    {
+      ret = FALSE;
+      goto out;
+    }
+
+  daemon = udisks_linux_mdraid_object_get_daemon (object);
+
+  if (!udisks_daemon_util_get_caller_uid_sync (daemon,
+                                               invocation,
+                                               NULL /* GCancellable */,
+                                               &caller_uid,
+                                               &caller_gid,
+                                               NULL,
+                                               error))
+    {
+      ret = FALSE;
+      goto out;
+    }
+
+  message = N_("Authentication is required to delete a RAID array");
+  action_id = "org.freedesktop.udisks2.manage-md-raid";
+  if (!udisks_daemon_util_check_authorization_sync_with_error (daemon,
+                                                               NULL,
+                                                               action_id,
+                                                               options,
+                                                               message,
+                                                               invocation,
+                                                               error))
+    {
+      ret = FALSE;
+      goto out;
+    }
+
+  member_devices = udisks_linux_mdraid_object_get_members (object);
+  raid_device = udisks_linux_mdraid_object_get_device (object);
+
+  if (teardown_flag)
+    {
+      message = N_("Authentication is required to modify the system configuration");
+      action_id = "org.freedesktop.udisks2.modify-system-configuration";
+      if (!udisks_daemon_util_check_authorization_sync_with_error (daemon,
+                                                                   NULL,
+                                                                   action_id,
+                                                                   options,
+                                                                   message,
+                                                                   invocation,
+                                                                   error))
+        {
+          ret = FALSE;
+          goto out;
+        }
+
+      if (raid_device)
+        {
+          /* The array is running, teardown its block device.
+          */
+          UDisksObject *block_object =
+            udisks_daemon_find_block_by_device_file (daemon,
+                                                     g_udev_device_get_device_file (raid_device->udev_device));
+          UDisksBlock *block = block_object? udisks_object_peek_block (block_object) : NULL;
+          if (block &&
+              !udisks_linux_block_teardown (block,
+                                            invocation,
+                                            options,
+                                            error))
+            {
+              g_clear_object (&block_object);
+              ret = FALSE;
+              goto out;
+            }
+
+          g_clear_object (&block_object);
+        }
+      else
+        {
+          /* The array is not running, remove the ChildConfiguration.
+          */
+          if (!udisks_linux_remove_configuration (udisks_mdraid_get_child_configuration (mdraid),
+                                                  error))
+            {
+              ret = FALSE;
+              goto out;
+            }
+        }
+    }
+
+  if (raid_device &&
+      !udisks_linux_mdraid_stop (mdraid,
+                                 invocation,
+                                 options,
+                                 error))
+    {
+      ret = FALSE;
+      goto out;
+    }
+
+  for (l = member_devices; l; l = l->next)
+    {
+      UDisksLinuxDevice *member_device = UDISKS_LINUX_DEVICE (l->data);
+      const gchar *device = g_udev_device_get_device_file (member_device->udev_device);
+      gchar *escaped_device = udisks_daemon_util_escape_and_quote (device);
+      gchar *error_message = NULL;
+      int status;
+
+      if (!udisks_daemon_launch_spawned_job_sync (daemon,
+                                                  UDISKS_OBJECT (object),
+                                                  "format-erase", caller_uid,
+                                                  NULL, /* cancellable */
+                                                  0,    /* uid_t run_as_uid */
+                                                  0,    /* uid_t run_as_euid */
+                                                  &status,
+                                                  &error_message,
+                                                  NULL, /* input_string */
+                                                  "wipefs -a %s",
+                                                  escaped_device))
+        {
+          g_set_error (error,
+                       UDISKS_ERROR,
+                       UDISKS_ERROR_FAILED,
+                       "Error wiping device: %s",
+                       error_message);
+          ret = FALSE;
+          g_free (escaped_device);
+          g_free (error_message);
+          goto out;
+        }
+
+      g_free (escaped_device);
+      g_free (error_message);
+    }
+
+  ret = TRUE;
+
+ out:
+  g_list_free_full (member_devices, g_object_unref);
+  g_clear_object (&raid_device);
+  return ret;
+}
+
+static gboolean
+handle_delete (UDisksMDRaid           *_mdraid,
+               GDBusMethodInvocation  *invocation,
+               GVariant               *options)
+{
+  GError *error = NULL;
+
+  if (!udisks_linux_mdraid_delete (_mdraid,
+                                     invocation,
+                                     options,
+                                     &error))
+    g_dbus_method_invocation_take_error (invocation, error);
+  else
+    udisks_mdraid_complete_delete (_mdraid, invocation);
+
+  return TRUE;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
 static void
 mdraid_iface_init (UDisksMDRaidIface *iface)
 {
@@ -1523,4 +1809,5 @@ mdraid_iface_init (UDisksMDRaidIface *iface)
   iface->handle_add_device = handle_add_device;
   iface->handle_set_bitmap_location = handle_set_bitmap_location;
   iface->handle_request_sync_action = handle_request_sync_action;
+  iface->handle_delete = handle_delete;
 }
